/*
	Copyright 2013 Anna Jezierska
	This function computes the gradient of Poisson-Gaussian data fidelity term 
	as described in
	
	[1] Anna Jezierska, Emilie Chouzenoux, Jean-Christophe Pesquet, and Hugues Talbot
	A primal-dual proximal splitting approach for restoring data corrupted with Poisson-Gaussian noise,
	International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
	Kyoto, 25-30 March, 2012.
	
	[2] Anna Jezierska, Emilie Chouzenoux, Jean-Christophe Pesquet, and Hugues Talbot
	A Convex Approach for Image Restoration with Exact Poisson-Gaussian Likelihood,
        submitted to SIAM Journal on Imaging Science, 2015
	
	[3] Anna Jezierska (2013). 
	Image Restoration in the presence of Poisson-Gaussian noise. 
	Ph.D. Thesis. Universit Paris-Est

	This software was developed by Anna Jezierska version of 09.2013
	(Anna.Jezierska@ibspan.waw.pl).
	
	If you use this software for research purposes, you should cite
	the aforementioned paper in any resulting publication.
	
	An example of compilation instruction:
	mex gradPGp.cpp LambertW/LambertW.cc
*/

#include "logpoissgauss.h"
#include <time.h>
#include "mex.h"
#include <pthread.h> 



#define maxaj(a, b) (((a) > (b)) ? (a) : (b))
#define minaj(a, b) (((a) < (b)) ? (a) : (b))

/*
--------------------------------------------------
STRUCTURE ST
--------------------------------------------------
gghx - pointer to the resulting vector gghx = 1-zeta(x). 
In general case x = Hx. Hence the gradient is given as (H^top * gghx).
xs - pointer to the orginal signal values
ys - pointer to the measured signal values
eTy - pointer to the table whose i-th element for i = {0, ..., dif} is defined as exp((y[i]-mean)*alpha/var)
alpha2var2 - constant equal to alpha^2/( 2 var )
mean - constant corresponding to mean of the Gaussian noise
var2 - constant corresponding to variance of the Gaussian noise
alpha - constant corresponding to scale parameter of the Poisson component of the noise
cf - constant equal to exp(- SQ(alpha)/(2.0*var)) 
delta - constant related to finite sums limits. It controls the number of coesfficients included into the finite sums.
nmax - size of the table exp_table_n
exp_table_n - pointer to the table whose n-th element for n = {1, ..., (nmax)} is defined as exp(- (n*alpha^2/var))
dif - size of the signal x (resp. y)
--------------------------------------------------
*/

typedef struct ST{
    double *gghx;
    double *xs;  
	double *ys; 
	double *eTy;
	double alpha2var2; 
	double mean;  
	double var2;   
	double alpha; 
	double cf;   
	double delta; 
	int nmax;
	double *exp_table_n;
	int dif;
}thestruct; 


/*
--------------------------------------------------
GRADIENT OF POISSON-GAUSSIAN term (no convolution kernel)
--------------------------------------------------
gghx = 1-zeta(x)
where zeta(x) = T/S = s(a,b-1)/s(a,b)
with s(a,b) = sum_0^inf pi(a,b,n)
--------------------------------------------------
pi(a,b,n) = a^n/n! exp(-(b-n)^2/2 var)
--------------------------------------------------
Recursive formular for computing the coefficients of infinte sums:
pi(a,b,n) = pi(a,b,n-1) * a/n * {exp ( b/var ) * exp ( -(n-1)/var ) * exp (-1/(2 var))}
Note that:
1. exp (-1/(2 var)) is a constant given by variable cf
2. exp ( b/var ) does not depend on estimated signal a and n. Hence it is precomputed and saved in table eTy
3. exp ( -n/var ) does not depend on estimated signal a and measuared signal b. 
   Hence it is precomputed and saved in table exp_table_n
--------------------------------------------------
Note that we can rewrite  pi(a,b,n) as:
pi(a,b,n) = aBYn(a,n) * expp(b,n)
and using the new variables rewrite:
T = sum_0^inf aBYn(a,n) * expp(b-1,n) = sum_0^inf aBYn(a,n) * expp(b,n+1)
S = sum_0^inf aBYn(a,n) * expp(b,n)
where
aBYn(a,n) = a^n/n!
expp(b,n) = exp(-(b-n)^2/(2 var))
aBYn(a,n+1) = aBYn(a,n) *a/(n+1)
expp(b,n+1) = expp(b,n) * exp ( b/var ) * exp ( -n/var ) * exp (-1/(2 var))
--------------------------------------------------
Infinite sum bounds [n_minus, n_plus]:
--------------------------------------------------
nstar = var W(a/sigma^2 exp (b/var) ) where W is a Lambert function
n_minus = nstar - delta
n_plus  = nstar + delta
where delta = CONST * sigma.
In this implementation CONST is set to 5. 
--------------------------------------------------
RECURSIVE ALGORITHM FOR COMPUTING THE GRADIENT OF POISSON-GAUSSIAN term (no convolution kernel)
--------------------------------------------------

FOR i = 1:dif DO:
    IF xs[i]>0
		1. Find nstar using "nstar = var W(xs[i]/sigma^2 exp (ys[i]/var) )"
		2. Find [n_minus, n_plus]
		3. Init S = expp(ys[i],0), T = expp(ys[i]-1,0); (with element for n=0)
		4. Init aBYn(xs[i],n_minus-1)
		5. Init expp(ys[i],n_minus)
		6. FOR n = n_minus: n_plus DO:
			6A. aBYn*= xs[i]/n;
			6B. S+=aBYn * expp;
			6C. expp *= exp ( b/var ) * exp ( -(n)/var ) * exp (-1/(2 var))
			6D. T+= aBYn * expp;
		END
		7. Set gghx[i] = 1-T/S
	ELSE
		1. We use a quadratic extension on negative orhant (-inf,0] -> 
		gradient is a linear function: gghx[i] = C1 xs[i] + C2
		where C1 = (1-exp(-1/var))*exp((2ys[i]-1)/var) ; C2 = 1-exp((1-2ys[i])/(2var))
END	
*/

void* gradPGpix( void* parm)
{

    thestruct* s = (thestruct *)parm;
	 
	double *gghx = (s)->gghx;
	double *xs = (s)->xs;
	double *ys = (s)->ys;
	double *eTy = (s)->eTy;
	double alpha2var2 = (s)->alpha2var2;
	double mean = (s)->mean;
	double var2 = (s)->var2;
	double alpha = (s)->alpha;
	double cf = (s)->cf;
	double delta = (s)->delta;
	int nmax = (s)->nmax;
	double *exp_table_n = (s)->exp_table_n;
	int dif = (s)->dif;
	double var = 0.5*var2;

	
	for (int i = 0; i <dif; i++)
	{
		
		if (xs[i] > 0 )
		{
			// Infinite sum bounds
			double lam = eTy[i]*xs[i]*alpha*alpha/var;
			double xstar =  LambertW(0, lam)*var/alpha/alpha;
			
			int lu = (int)ceil(xstar + delta);
			int ld = (int)floor(xstar - delta);
			if (ld < 1) ld = 1;
			
			// Init S and T with zero element of infinite sums
			double expp = exp(- SQ((ys[i]-mean))/(var2));
			double S = expp;
			double ycf = eTy[i] *cf;
			double T = S*ycf;
			
			//Init expp(ys[i],n_minus)
			expp = exp(- SQ(ys[i]-mean-alpha*ld)/(var2));
			
			//Init aBYn(xs[i],n_minus-1)
			double xBYn = 1;
			
			for (int n = 1; n <ld; ++n)
			{
				//xBYn *=  (xs/(double)n);
				xBYn *= xs[i]; // x^n / n!
				xBYn /= n;
			}
		    
			//Add to S and T all significant elements of infinite sum 
			int loopend = minaj(lu,nmax);
			
			for (int n = ld; n < loopend; ++n)
			{
			   // double nn = n;
				//xBYn *= xs/(double)n; // x^n / n!
				
				xBYn *= xs[i]; // x^n / n!
				xBYn /= n;				
				S+= (xBYn*expp);
				
				//expp *= (*(eTn)) * ycf;
				
				expp *= exp_table_n[n]; //(*(eTn));
				expp *= ycf;
				
				T+= (xBYn*expp);
				
				
				//++eTn;                    
			}
			 
			if (lu > nmax) 
			{
				double coef = SQ(alpha)/var;
				for (int n = nmax+1; n < lu; ++n)
				{
			   
					xBYn *= xs[i]/(double)n; // x^n / n!
				
					S+= xBYn*expp;
				
					double eTnn = exp(- (n*coef));
					expp *= eTnn * ycf;
				
					T+= xBYn*expp;
				
						   
				}
				//printf("Sum bounds exceed size of array [%d %d], nmax: %d, y %f, x %f", ld, lu,nmax, y[i], x[i] );
				//return;
			}
			//Set result
			if (S>0)
			 gghx[i] = (1- T/S);
			else
			 gghx[i] = 0;
		}
			
		else
		{
       	
			double C1 = (1-exp(-1/var)) * exp((2*(ys[i]-mean) -1)/(var));
			double C2 = 1 - exp(( ys[i]*2 -1)/(var2));
			gghx[i] =   C1 * xs[i] + C2 ;
		}
		
	}

}
/*------------------------------------
THIS FUNCTION ALLOWS FOR PARALLELIZATION OF THE CODE
IT CALLS FUNCTION gradPGpix FOR BLOCKS OF PIXELS
------------------------------------*/
void gradPGp     (double* ghx,
                 double*  x,
                 double*  y,
	         double*  exp_table_y,
	         double*  exp_table_n,
                 double var,
                 double alpha,
                 double mean,   
                 int size,
                 int nmax,
                 double tolerance
                 )
{
   
        double sigma = sqrt (var);
       // double tolerance = 5.0;
	int NO_PROC = 1; //maxaj(1,3*sysconf( _SC_NPROCESSORS_ONLN ) / 4);
   
	pthread_t             *thread; // for linux
	
	thread = (pthread_t *) malloc(NO_PROC * sizeof(*thread));
	
	int j = 0;
	int dif = (int)(size/NO_PROC);
	double sum_bounds  = tolerance*ceil(sigma/alpha)+1.0;
	int del =  int(sum_bounds);// here we define summations bounds
	//printf("dif: %d", dif );
	
	double SQav = SQ(alpha)/(2.0*var);
	double cff = exp(-SQav);
	
    for (int i = 0; i < NO_PROC ; ++i)
    {
	    void *status = 0;

		thestruct *st = NULL;
	    st = (thestruct *) malloc(sizeof(thestruct));
		
		st->xs = x + i*dif;
                st->ys = y+i*dif;
		st->eTy = exp_table_y +i*dif;
		st->gghx = ghx + i*dif;
		st->exp_table_n = exp_table_n;
		st->var2 = 2.0*var;
		st->alpha2var2 = SQav;
		st->cf =  cff; // correction factor
		st->delta = del;
		st->nmax = nmax;
		st->mean = mean;
		st->alpha = alpha;
		st->dif = dif;
		if (i == NO_PROC -1)
		st->dif = size - i*dif;
		//printf("dif: %d \n", i*dif );
         
		//_beginthread((void(*)(void*))gradPGpix,0,(void*)&st); //for windows
		

		pthread_create(&thread[i], NULL, gradPGpix, (void *) st);
				
       
    }
	for (int i = 0; i < NO_PROC ; ++i)
	{
		//pthread_join(thread[NO_PROC-1], NULL);
		pthread_join(thread[i], NULL);
	}
	
     
}


/* the gateway function */
void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[])
{
  double *y,*x, *exp_table_y, *exp_table_n,*ghx;
  double var, mean, alpha, tolerance ;
 // double tolerance = 5.0;
  
  size_t size, nmax;

  
  /*  check for proper number of arguments */
  /* NOTE: You do not need an else statement when using mexErrMsgIdAndTxt
     within an if statement, because it will never get to the else
     statement if mexErrMsgIdAndTxt is executed. (mexErrMsgIdAndTxt breaks you out of
     the MEX-file) */
  if(nrhs!=8) 
    mexErrMsgIdAndTxt( "MATLAB:xtimesy:invalidNumInputs",
            "Five inputs required: x - vector (orginal signal); y - vector of observations; exp_table_y - vector defined as: exp_table_y[i] = exp(alpha/var (y[i] - mean) ); exp_table_n - vector defined as: exp_table_n[i] = exp(-n alpha^2/var); var - scalar (Gaussian noise variance; mean - scalar (Gaussian noise mean); alpha- scalar (scaling parameter of Poisson noise)"
			);
  if(nlhs!=1) 
    mexErrMsgIdAndTxt( "MATLAB:xtimesy:invalidNumOutputs",
            "One output required.");
  
  /* check to make sure the thd input argument is a scalar */
  if( !mxIsDouble(prhs[4]) || mxIsComplex(prhs[4]) ||
      mxGetN(prhs[4])*mxGetM(prhs[4])!=1 ) {
    mexErrMsgIdAndTxt( "MATLAB:gradPG:varNotScalar",
            "Input var must be a scalar.");
  }
  /* check to make sure the mean input argument is a scalar */
  if( !mxIsDouble(prhs[5]) || mxIsComplex(prhs[5]) ||
      mxGetN(prhs[5])*mxGetM(prhs[5])!=1 ) {
    mexErrMsgIdAndTxt( "MATLAB:gradPG:meanNotScalar",
            "Input mean must be a scalar.");
  }
  
  /* check to make sure the alpha input argument is a scalar */
  if( !mxIsDouble(prhs[6]) || mxIsComplex(prhs[6]) ||
      mxGetN(prhs[6])*mxGetM(prhs[6])!=1 ) {
    mexErrMsgIdAndTxt( "MATLAB:gradPG:alphaNotScalar",
            "Input alpha must be a scalar.");
  }

  /*  get the scalar input alpha */
  tolerance = mxGetScalar(prhs[7]);


  /*  get the scalar input alpha */
  alpha = mxGetScalar(prhs[6]);
  
  /*  get the scalar input mean */
  mean = mxGetScalar(prhs[5]);
  
  /*  get the scalar input var */
  var = mxGetScalar(prhs[4]);
  
  /*  create a pointer to the input matrix y */
  exp_table_n = mxGetPr(prhs[3]);
  
  /*  create a pointer to the input matrix y */
  exp_table_y = mxGetPr(prhs[2]);
  
  /*  create a pointer to the input matrix y */
  y = mxGetPr(prhs[1]);
  
   /*  create a pointer to the input matrix x */
  x = mxGetPr(prhs[0]);
  
  /*  get the dimensions of the matrix input y */
  size = maxaj(mxGetM(prhs[0]),mxGetN(prhs[0]));
  if ((maxaj(mxGetM(prhs[1]),mxGetN(prhs[1])) != size) 
    || (maxaj(mxGetM(prhs[2]),mxGetN(prhs[2])) != size)  
    || (minaj(mxGetM(prhs[1]),mxGetN(prhs[1])) != 1) 
    ||  (minaj(mxGetM(prhs[0]),mxGetN(prhs[0])) != 1) 
	||  (minaj(mxGetM(prhs[2]),mxGetN(prhs[2])) != 1) )
  {
      mexErrMsgIdAndTxt( "MATLAB:gradPG:x , y , exp_table_y NotVector",
            "Input x,y, exp_table_y must have the same size and be vectors.");
  }
  
  nmax = maxaj(mxGetM(prhs[3]),mxGetN(prhs[3]));
  if (minaj(mxGetM(prhs[3]),mxGetN(prhs[3])) != 1)
  {
      mexErrMsgIdAndTxt( "MATLAB: gradPG : exp_table_n NotVector",
            "Input exp_table_n must be a vector.");
  }
 
  
  /*  set the output pointer to the output matrix */
  plhs[0] = mxCreateDoubleMatrix( size, 1, mxREAL);
  
  /*  create a C pointer to a copy of the output matrix */
  ghx = mxGetPr(plhs[0]);
  
  
  
  /*  call the C subroutine */

  //int t1 = clock();
  gradPGp(ghx, x,y,exp_table_y, exp_table_n, var,alpha,mean,size, nmax,tolerance);
  //sleep(100);
  //int t2 = clock();
  //printf ("It took me %f seconds [%d, %d] .\n",((float)(t2-t1))/CLOCKS_PER_SEC, t1, t2);
}
